
#if defined(__AVR_ATmega32U4__) || defined(TEENSY20)
  #define PROMICRO
#endif
#define ACC 1
#define GYRO 1
#define QUADX

#define NUMBER_MOTOR     4

//#define EXT_MOTOR_64KHZ
//#define EXT_MOTOR_32KHZ
//#define EXT_MOTOR_16KHZ
#define EXT_MOTOR_8KHZ  // i recommend this one for a smooth flight


/****************************    Motor minthrottle    *******************************/
  /* Set the minimum throttle command sent to the ESC (Electronic Speed Controller)
     This is the minimum value that allow motors to run at a idle speed  */
#define MAXTHROTTLE 1900
#define MINTHROTTLE 1050 // for brushed ESCs like ladybird
/****************************    Mincommand          *******************************/
  /* this is the value for the ESCs when they are not armed
     in some cases, this value must be lowered down to 900 for some specific ESCs, otherwise they failed to initiate */
#define MINCOMMAND  1000
#define HWPWM6

#define MINCHECK 1100
#define MAXCHECK 1900

/**********************************    I2C speed   ************************************/
  //#define I2C_SPEED 100000L     //100kHz normal mode, this value must be used for a genuine WMP
#define I2C_SPEED 400000L   //400kHz fast mode, it works only with some WMP clones

#define I2C_PULLUPS_DISABLE        PORTD &= ~(1<<0); PORTD &= ~(1<<1);
#define MPU6050
#define MPU6050_ADDRESS     0x68 // address pin AD0 low (GND), default for FreeIMU v0.4 and InvenSense evaluation board
//#define MPU6050_LPF_256HZ
//#define MPU6050_LPF_188HZ
//#define MPU6050_LPF_98HZ
#define MPU6050_LPF_42HZ
//#define MPU6050_LPF_20HZ
//#define MPU6050_LPF_10HZ
//#define MPU6050_LPF_5HZ

//MPU6050 Gyro LPF setting
#if defined(MPU6050_LPF_256HZ) || defined(MPU6050_LPF_188HZ) || defined(MPU6050_LPF_98HZ) || defined(MPU6050_LPF_42HZ) || defined(MPU6050_LPF_20HZ) || defined(MPU6050_LPF_10HZ) || defined(MPU6050_LPF_5HZ)
  #if defined(MPU6050_LPF_256HZ)
    #define MPU6050_DLPF_CFG   0
  #endif
  #if defined(MPU6050_LPF_188HZ)
    #define MPU6050_DLPF_CFG   1
  #endif
  #if defined(MPU6050_LPF_98HZ)
    #define MPU6050_DLPF_CFG   2
  #endif
  #if defined(MPU6050_LPF_42HZ)
    #define MPU6050_DLPF_CFG   3
  #endif
  #if defined(MPU6050_LPF_20HZ)
    #define MPU6050_DLPF_CFG   4
  #endif
  #if defined(MPU6050_LPF_10HZ)
    #define MPU6050_DLPF_CFG   5
  #endif
  #if defined(MPU6050_LPF_5HZ)
    #define MPU6050_DLPF_CFG   6
  #endif
#else
    //Default settings LPF 256Hz/8000Hz sample
    #define MPU6050_DLPF_CFG   0
#endif

#define LEDPIN_PINMODE             //
#define LEDPIN_TOGGLE              PIND |= 1<<5;     //switch LEDPIN state (Port D5)
#define LEDPIN_OFF                 PORTD |= (1<<5);
#define LEDPIN_ON                  PORTD &= ~(1<<5);

#if 0 // NOT USED
#if defined(MPU6050)
  #define GYRO_SCALE (4 / 16.4 * PI / 180.0 / 1000000.0)   //MPU6050 and MPU3050   16.4 LSB/(deg/s) and we ignore the last 2 bits
#endif
#endif

#define ACC_ORIENTATION(X, Y, Z)  {imu.accADC[ROLL]  =  -X; imu.accADC[PITCH]  = -Y; imu.accADC[YAW]  = Z;}
#define GYRO_ORIENTATION(X, Y, Z) {imu.gyroADC[ROLL] = Y; imu.gyroADC[PITCH] =  -X; imu.gyroADC[YAW] = -Z;}

#if 0 // NOT USED
#define ACCRESO 4096  //acceleratin resolution
#define ACC_1G 512
#define ACCZ_25deg   (int16_t)(ACC_1G * 0.90631) // 0.90631 = cos(25deg) (cos(theta) of accZ comparison)
#define ACC_VelScale (9.80665f / 10000.0f / ACC_1G)
#define ACC_LPF_FACTOR 4 // that means a LPF of 16
#endif

enum rc {
  ROLL,
  PITCH,
  YAW,
  THROTTLE,
  AUX1,
  AUX2,
  AUX3,
  AUX4,
  AUX5,
  AUX6,
  AUX7,
  AUX8
};

typedef struct {
  int16_t  accSmooth[3];
  int16_t  gyroData[3];
  int16_t  magADC[3];
  int16_t  gyroADC[3];
  int16_t  accADC[3];
} imu_t;

int16_t  i2c_errors_count = 0;
uint8_t rawADC[6];
static uint32_t neutralizeTime = 0;

uint32_t currentTime = 0;
uint16_t previousTime = 0;
uint16_t cycleTime = 0;     // this is the number in micro second to achieve a full loop, it can differ a little and is taken into account in the PID loop

int16_t motor[8];
uint8_t PWM_PIN[8] = {9,10,5,6,11,13};   //for a quad+: rear,right,left,front

imu_t imu;

void blinkLED(uint8_t num, uint8_t ontime,uint8_t repeat) {
  uint8_t i,r;
  for (r=0;r<repeat;r++) {
    for(i=0;i<num;i++) {
      LEDPIN_TOGGLE; // switch LEDPIN state
      delay(ontime);
    }
    delay(60); //wait 60 ms
  }
}

void i2c_init(void) {
  I2C_PULLUPS_DISABLE
  TWSR = 0;                                    // no prescaler => prescaler = 1
  TWBR = ((F_CPU / I2C_SPEED) - 16) / 2;       // change the I2C clock rate
  TWCR = 1<<TWEN;                              // enable twi module, no interrupt
}

void i2c_rep_start(uint8_t address) {
  TWCR = (1<<TWINT) | (1<<TWSTA) | (1<<TWEN) ; // send REPEAT START condition
  waitTransmissionI2C();                       // wait until transmission completed
  TWDR = address;                              // send device address
  TWCR = (1<<TWINT) | (1<<TWEN);
  waitTransmissionI2C();                       // wail until transmission completed
}

void i2c_stop(void) {
  TWCR = (1 << TWINT) | (1 << TWEN) | (1 << TWSTO);
  //  while(TWCR & (1<<TWSTO));                // <- can produce a blocking state with some WMP clones
}

void i2c_write(uint8_t data ) {
  TWDR = data;                                 // send data to the previously addressed device
  TWCR = (1<<TWINT) | (1<<TWEN);
  waitTransmissionI2C();
}

uint8_t i2c_read(uint8_t ack) {
  TWCR = (1<<TWINT) | (1<<TWEN) | (ack? (1<<TWEA) : 0);
  waitTransmissionI2C();
  uint8_t r = TWDR;
  if (!ack) i2c_stop();
  return r;
}

uint8_t i2c_readAck() {
  return i2c_read(1);
}

uint8_t i2c_readNak(void) {
  return i2c_read(0);
}

void waitTransmissionI2C() {
  uint16_t count = 255;
  while (!(TWCR & (1<<TWINT))) {
    count--;
    if (count==0) {              //we are in a blocking state => we don't insist
      TWCR = 0;                  //and we force a reset on TWINT register
      neutralizeTime = micros(); //we take a timestamp here to neutralize the value during a short delay
      i2c_errors_count++;
      break;
    }
  }
}

size_t i2c_read_to_buf(uint8_t add, void *buf, size_t size) {
  i2c_rep_start((add<<1) | 1);  // I2C read direction
  size_t bytes_read = 0;
  uint8_t *b = (uint8_t*)buf;
  while (size--) {
    /* acknowledge all but the final byte */
    *b++ = i2c_read(size > 0);
    /* TODO catch I2C errors here and abort */
    bytes_read++;
  }
  return bytes_read;
}

size_t i2c_read_reg_to_buf(uint8_t add, uint8_t reg, void *buf, size_t size) {
  i2c_rep_start(add<<1); // I2C write direction
  i2c_write(reg);        // register selection
  return i2c_read_to_buf(add, buf, size);
}

/* transform a series of bytes from big endian to little
   endian and vice versa. */
void swap_endianness(void *buf, size_t size) {
  /* we swap in-place, so we only have to
  * place _one_ element on a temporary tray
  */
  uint8_t tray;
  uint8_t *from;
  uint8_t *to;
  /* keep swapping until the pointers have assed each other */
  for (from = (uint8_t*)buf, to = &from[size-1]; from < to; from++, to--) {
    tray = *from;
    *from = *to;
    *to = tray;
  }
}

void i2c_getSixRawADC(uint8_t add, uint8_t reg) {
  i2c_read_reg_to_buf(add, reg, &rawADC, 6);
}

void i2c_writeReg(uint8_t add, uint8_t reg, uint8_t val) {
  i2c_rep_start(add<<1); // I2C write direction
  i2c_write(reg);        // register selection
  i2c_write(val);        // value to write in register
  i2c_stop();
}

uint8_t i2c_readReg(uint8_t add, uint8_t reg) {
  uint8_t val;
  i2c_read_reg_to_buf(add, reg, &val, 1);
  return val;
}

void Gyro_init() {
  TWBR = ((F_CPU / 400000L) - 16) / 2; // change the I2C clock rate to 400kHz
  i2c_writeReg(MPU6050_ADDRESS, 0x6B, 0x80);             //PWR_MGMT_1    -- DEVICE_RESET 1
  delay(5);
  i2c_writeReg(MPU6050_ADDRESS, 0x6B, 0x03);             //PWR_MGMT_1    -- SLEEP 0; CYCLE 0; TEMP_DIS 0; CLKSEL 3 (PLL with Z Gyro reference)
  i2c_writeReg(MPU6050_ADDRESS, 0x1A, MPU6050_DLPF_CFG); //CONFIG        -- EXT_SYNC_SET 0 (disable input pin for data sync) ; default DLPF_CFG = 0 => ACC bandwidth = 260Hz  GYRO bandwidth = 256Hz)
  i2c_writeReg(MPU6050_ADDRESS, 0x1B, 0x18);             //GYRO_CONFIG   -- FS_SEL = 3: Full scale set to 2000 deg/sec
  // enable I2C bypass for AUX I2C
}

void Gyro_getADC () {
  i2c_getSixRawADC(MPU6050_ADDRESS, 0x43);
  GYRO_ORIENTATION( ((rawADC[0]<<8) | rawADC[1])>>2 , // range: +/- 8192; +/- 2000 deg/sec
                    ((rawADC[2]<<8) | rawADC[3])>>2 ,
                    ((rawADC[4]<<8) | rawADC[5])>>2 );
}

void ACC_init () {
  i2c_writeReg(MPU6050_ADDRESS, 0x1C, 0x10);             //ACCEL_CONFIG  -- AFS_SEL=2 (Full Scale = +/-8G)  ; ACCELL_HPF=0   //note something is wrong in the spec.
  //note: something seems to be wrong in the spec here. With AFS=2 1G = 4096 but according to my measurement: 1G=2048 (and 2048/8 = 256)
  //confirmed here: http://www.multiwii.com/forum/viewtopic.php?f=8&t=1080&start=10#p7480
}

void ACC_getADC () {
  i2c_getSixRawADC(MPU6050_ADDRESS, 0x3B);
  ACC_ORIENTATION( ((rawADC[0]<<8) | rawADC[1])>>3 ,
                   ((rawADC[2]<<8) | rawADC[3])>>3 ,
                   ((rawADC[4]<<8) | rawADC[5])>>3 );
}

void initSensors() {
  delay(200);
  //POWERPIN_ON;
  delay(100);
  i2c_init();
  delay(100);
  Gyro_init();
  ACC_init();
}

/**************************************************************************************/
/************  Writes the Motors values to the PWM compare register  ******************/
/**************************************************************************************/
void writeMotors() { // [1000;2000] => [125;250]
   #if !(NUMBER_MOTOR == 4)
   #error "only 4 motors allowed"
   #endif
   
   #if !defined(PROMICRO)//
   #error "only arduino leonardo, go tools -> board and set leonardo"
   #endif
   
   #if defined(PROMICRO)
   uint16_t Temp2;
   Temp2 = motor[3] - 1000;
   #if defined(EXT_MOTOR_64KHZ)
   OCR1A = (motor[0] - 1000) >> 2; // max = 255
   OCR1B = (motor[1] - 1000) >> 2;
   OCR3A = (motor[2] - 1000) >> 2;
   Temp2 = Temp2 >> 2;
   TC4H = Temp2 >> 8;
   OCR4D = Temp2 & 0xFF;         //  pin 6
   #elif defined(EXT_MOTOR_32KHZ)
   OCR1A = (motor[0] - 1000) >> 1; // max = 511
   OCR1B = (motor[1] - 1000) >> 1;
   OCR3A = (motor[2] - 1000) >> 1;
   Temp2 = Temp2 >> 1;
   TC4H = Temp2 >> 8;
   OCR4D = Temp2 & 0xFF;         //  pin 6
   #elif defined(EXT_MOTOR_16KHZ)
   OCR1A = motor[0] - 1000; //  pin 9
   OCR1B = motor[1] - 1000; //  pin 10
   OCR3A = motor[2] - 1000; //  pin 5
   TC4H = Temp2 >> 8;
   OCR4D = Temp2 & 0xFF;         //  pin 6
   #elif defined(EXT_MOTOR_8KHZ)
   OCR1A = (motor[0]-1000) << 1; //  pin 9
   OCR1B = (motor[1]-1000) << 1; //  pin 10
   OCR3A = (motor[2]-1000) << 1; //  pin 5
   TC4H = Temp2 >> 8;
   OCR4D = Temp2 & 0xFF;         //  pin 6
   #endif
   #endif
}

/**************************************************************************************/
/************          Writes the mincommand to all Motors           ******************/
/**************************************************************************************/
void writeAllMotors(int16_t mc) {   // Sends commands to all motors
  for (uint8_t i =0;i<NUMBER_MOTOR;i++) {
    motor[i]=mc;
  }
  writeMotors();
}

/**************************************************************************************/
/************        Initialize the PWM Timers and Registers         ******************/
/**************************************************************************************/
void initOutput() {
  /****************            mark all PWM pins as Output             ******************/
  for(uint8_t i=0;i<NUMBER_MOTOR;i++) {
    pinMode(PWM_PIN[i],OUTPUT);
  }
  
  #if defined(PROMICRO)
  TCCR1A = (1<<WGM11);
  TCCR1B = (1<<WGM13) | (1<<WGM12) | (1<<CS10);
  
  #if defined(EXT_MOTOR_64KHZ)
  delayMicroseconds(9);
  #elif defined(EXT_MOTOR_32KHZ)
  delayMicroseconds(16);
  #elif defined(EXT_MOTOR_16KHZ)
  delayMicroseconds(32);
  #elif defined(EXT_MOTOR_8KHZ)
  delayMicroseconds(64);
  #endif

  TCCR3A = (1<<WGM31);
  TCCR3B = (1<<WGM33) | (1<<WGM32) | (1<<CS30);
  
  #if defined(EXT_MOTOR_64KHZ)
  ICR1   = 0x00FF; // TOP to 255;
  ICR3   = 0x00FF; // TOP to 255;
  delayMicroseconds(14);
  TC4H = 0x00;
  OCR4C = 0xFF; // phase and frequency correct mode & top to 255
  TCCR4B = (1<<CS40);             // prescaler to 1
  #elif defined(EXT_MOTOR_32KHZ)
  ICR1   = 0x01FF; // TOP to 511;
  ICR3   = 0x01FF; // TOP to 511;
  delayMicroseconds(14);
  TC4H = 0x01;
  OCR4C = 0xFF; // phase and frequency correct mode & top to 511
  TCCR4B = (1<<CS40);             // prescaler to 1
  #elif defined(EXT_MOTOR_16KHZ)
  ICR1   = 0x03FF; // TOP to 1023;
  ICR3   = 0x03FF; // TOP to 1023;
  delayMicroseconds(45);
  TC4H = 0x03;
  OCR4C = 0xFF; // phase and frequency correct mode & top to 1023
  TCCR4B = (1<<CS40);             // prescaler to 1
  #elif defined(EXT_MOTOR_8KHZ)
  ICR1   = 0x07FF; // TOP to 2046;
  ICR3   = 0x07FF; // TOP to 2046;
  delayMicroseconds(105);
  TC4H = 0x3;
  OCR4C = 0xFF; // phase and frequency correct mode
  TCCR4B = (1<<CS41);             // prescaler to 2
  #else
  #error only 64khz to 8khz on 32u4 device
  #endif
  
  TCCR1A |= _BV(COM1A1); // connect pin 9 to timer 1 channel A
  TCCR1A |= _BV(COM1B1); // connect pin 10 to timer 1 channel B
  TCCR3A |= _BV(COM3A1); // connect pin 5 to timer 3 channel A
  
  TCCR4D = 0;
  TCCR4C |= (1<<COM4D1)|(1<<PWM4D); // connect pin 6 to timer 4 channel D
  #endif

  writeAllMotors(MINCOMMAND);
  delay(300);
}

void setup()
{
  Serial.begin(9600);
  initOutput();
  initSensors();
}

void loop()
{
  static uint32_t rcTime  = 0;
  
  if (currentTime > rcTime ) { // 50Hz
    // RC loop
    rcTime = currentTime + 20000;
  } else {
    // Not RC loop
  }

  currentTime = micros();
  cycleTime = currentTime - previousTime;
  previousTime = currentTime;
  motor[0] = MINCOMMAND;
  motor[1] = MINCOMMAND;
  motor[2] = MINCOMMAND;
  motor[3] = MINCOMMAND+50;
  writeMotors();
  Serial.print("Motor [0]=");
  Serial.print(motor[0]);
  Serial.print(", [1]=");
  Serial.print(motor[1]);
  Serial.print(", [2]=");
  Serial.print(motor[2]);
  Serial.print(", [3]=");
  Serial.println(motor[3]);
  #if 0
  Serial.print("Gyro : ROLL=");
  Serial.print(imu.gyroADC[ROLL]);
  Serial.print(", PITCH=");
  Serial.print(imu.gyroADC[PITCH]);
  Serial.print(", YAW=");
  Serial.println(imu.gyroADC[YAW]);
  #endif
  #if 0
  Serial.print("ACC : ROLL=");
  Serial.print(imu.accADC[ROLL]);
  Serial.print(", PITCH=");
  Serial.print(imu.accADC[PITCH]);
  Serial.print(", YAW=");
  Serial.println(imu.accADC[YAW]);
  #endif
}

